/********************************************************
 * File: smparse.c (Listing 8)
 * Title: Parser and main()
 * Author: Allen Holub
 * Compiler: Turbo C 2.0, MSC 5.1
 ********************************************************/

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include "smtok.h"      /* contains definition for "token" enumerated type */
#include "smproto.h"    /* prototypes for functions in the current file    */

/* SMPARSE.C    This file contains the parser for a small state-machine
 *              description language. A simple "recursive" descent parser
 *              is used. I'm putting recursive in quotes because there's
 *              actually no recursion necessary with the current grammar
 *              ---I've replaced all the tail recursion with loops.
 *              This is a straightforward implementation of the grammar,
 *              in the article---one subroutine per left-hand side. The
 *              subroutines have the same names as the nonterminals they
 *              implement.
 *
 *              Invoke with:   smparse < input_file >output_file
 *
 *              Exit status is the number of errors encountered in the parse.
 */

#ifdef DEBUG
#define D(x) x
#else
#define D(x)
#endif

#define  SUCCESS NO_TOKEN
#define  FAILURE NO_TOKEN2
#define  CCL     NO_TOKEN3

#define MAX_ERR 32      /* Both maximum number of errors before abort and
                         * exit status when abort occurs.
                         */

static FILE  *Input     = stdin;        /* input stream         */
static token Lookahead  = NO_TOKEN;
#define match(tok) ((tok) == (Lookahead==NO_TOKEN ? advance() : Lookahead))
#define advance()  (Lookahead = gtok(Input))

#define MAXSTATES 128
#define MAXCOLS   128
static int   Statetab[ MAXSTATES ][ MAXCOLS ];
static char *Acceptab[ MAXSTATES ];
static int   State = 0;         /* current state number                     */
static int   Maxstate ;         /* maximum state number seen so far         */

                                /* These are for statistical purposes only: */
static int   Mincol;            /* Minimum column index seen so far.        */
static int   Maxcol;            /* Maximum column index seen so far.        */
static int   Error_count = 0;   /* running count of parse errors            */

extern char yytext[];   /* from smlex.c: lexeme for current token           */
extern int  yylineno;   /* from smlex.c: line number for current token      */

#ifndef max
#define max(a,b) ((a) > (b) ? (a) : (b))
#define min(a,b) ((a) < (b) ? (a) : (b))
#endif

/*----------------------------------------------------------------------*/

void    error( char *fmt, ... )  /* like printf, but adds line number and */
{                                /* goes to standard error.               */
    va_list     args;
    extern int  yylineno;       /* from smlex.c */

    va_start( args,   fmt );
    fprintf ( stderr, "sm (line %3d near \"%s\"): ", yylineno, yytext );
    vfprintf( stderr, fmt, args );
    va_end  ( args );

    if( ++Error_count >= MAX_ERR )
        exit( MAX_ERR );
}

static char *tok_to_str( token tok )
{
    return  (tok == ACCEPT              ) ? "\"accept\"" :
            (tok == CHARACTER           ) ? "char." :
            (tok == CLOSE_BRACKET       ) ? "\"]\"" :
            (tok == CLOSE_PAREN         ) ? "\")\"" :
            (tok == CODE                ) ? "\"code\"" :
            (tok == COLON               ) ? "\":\"" :
            (tok == DASH                ) ? "\"-\"" :
            (tok == GOTO                ) ? "\"goto\"" :
            (tok == LOOP                ) ? "\"loop\"" :
            (tok == NUMBER              ) ? "number" :
            (tok == ON                  ) ? "\"on\"" :
            (tok == OPEN_BRACKET        ) ? "\"[\"" :
            (tok == OPEN_PAREN          ) ? "\"(\"" :
            (tok == STATE               ) ? "\"state\"" :
                                            "***UNKNOWN*** ";
}

static int expected( token current, ... )
{
    /* arguments are a 0-terminated list of tokens. Return true if the
     * current lookahead token matches any of the ones in the list.
     */

    va_list     args;
    token       next;

    va_start     (args, current);
    next = va_arg(args, token);

    error( "Expected %s", tok_to_str( current ) );

    while( next )
    {
        current = next;
        next    = va_arg(args, token);

        fprintf( stderr, next ? (", %s") : (" or %s"), tok_to_str(current) );
    }

    fprintf( stderr, ". Skipping state.\n" );

    while( !match( STATE ) && !match(END_OF_FILE) )
        advance();

    va_end( args );
    return( FAILURE );
}

/*----------------------------------------------------------------------*/

int checkstate( int state )
{
    /* Check to see if the incoming state is valid and update various
     * statistics.
     */

    if( state > MAXSTATES )
    {
        error("State number too large, %d max\n", MAXSTATES );
        state = 0;
    }
    else if( state < 0 )
    {
        error("value of ON statement too small, must be > 0\n" );
        state = 0;
    }
    else
        Maxstate = max( state, Maxstate );

    return state;
}

int checkcol( int col )
{
    /* Check to see if the incoming argument is a valid column index and
     * update various statistics.
     */

    Mincol = min( col, Mincol );
    Maxcol = max( col, Maxcol );
    if( col > MAXCOLS )
    {
        error("value of ON statement too large, %d max\n", MAXCOLS );
        col = 0;
    }
    if( col < 0 )
    {
        error("value of ON statement too small, must be >0\n" );
        col = 0;
    }
    return col;
}

/*======================================================================*/

void    description()
{
    while( !match(END_OF_FILE) )
        state();
}

int     state()
{
    if( !match(STATE) )
        return expected( STATE, 0 );
    else
    {
        advance();
        if( !match( NUMBER) )
            return expected( NUMBER, 0 );
        else
        {
            State = checkstate( atoi(yytext) );
            advance();
        }

        if( match( COLON ) ) advance();
        else                 error( "Inserted missing colon\n" );

        return body();
    }
}

/*----------------------------------------------------------------------*/

int     body()
{
    while( transition() == SUCCESS  &&  !match(END_OF_FILE) )
        ;
}

/*----------------------------------------------------------------------*/

int     transition()
{
    int nextstate;

    switch( Lookahead )
    {
    case LOOP:
        advance();
        return on_val( State );

    case GOTO:
        advance();
        if( !match(NUMBER) )
            return expected( NUMBER, 0 );

        nextstate = atoi( yytext );
        advance();
        return on_val( nextstate );

    case ACCEPT:
        advance();
        if( !match(CODE) )
            return expected( CODE, 0 );

        Acceptab[ State ] = strdup( yytext );
        advance();
        return SUCCESS;

    default:
        if( !match(STATE) )
            return expected(STATE, LOOP, GOTO, ACCEPT, 0);
    }
}
/*----------------------------------------------------------------------*/
int     on_val( int nextstate )
{
    int i;

    if( match(ON) ) advance();
    else                error( "Inserted missing \"on\" statement\n" );

    if( (i = value()) == FAILURE )
        return FAILURE;
    else
    {
        if( i != CCL )
            addstate( State, i, nextstate );
        else
            while( (i = value()) != FAILURE )
                addstate( State, i, nextstate );
    }
    return SUCCESS;
}
/*----------------------------------------------------------------------*/
void addstate( row_index, col_index, contents )
{
    D( printf("Statetab[%d]['%c'] = %d\n", row_index, col_index, contents ); )

    Statetab[ row_index ][ col_index ] = contents ;
}
/*----------------------------------------------------------------------*/
int     value()
{
    /* Generally returns a value to use for a column index or FAILURE on
     * a syntax error. If it returns CCL, then subsequent value() calls will
     * return subsequent elements of the range until there are no such ele-
     * ments, in which case FAILURE is returned. checkcol() is called for each 
     * returned value. Both the value and val productions are handled here.
     */

    int    i, rval;
    int    negative;
    char   *p;
    static unsigned char *last_ccl = NULL, cclbuf[ MAXCOLS ];

    if( last_ccl )
    {
        while( !*last_ccl  && last_ccl < (cclbuf + MAXCOLS) )
            ++last_ccl;

        if( last_ccl < cclbuf + MAXCOLS )
            return last_ccl++ - cclbuf;
        else
        {
            last_ccl = NULL;    /* no more values */
            return FAILURE;
        }
    }

    if( negative = match( NOT ) )
    {
        if( match(GOTO) || match(STATE) || match(LOOP) || match(ACCEPT) )
        {
            /* It's a literal !, not a NOT token. */
            negative = 0;
        }
        else
            advance();
    }

    if( (i = literal()) != FAILURE )
        return i;

    switch( Lookahead )
    {
    case CLOSE_BRACKET:
    case CLOSE_PAREN:
    case COLON:
    case DASH:
    case NOT:
    case OPEN_PAREN:    rval = checkcol( *yytext );
                        advance();
                        break;
    case CHARACTER:
    case NUMBER:        rval = char_or_num();
                        break;

    case OPEN_BRACKET:  advance();
                        if( match(GOTO) || match(STATE) || match(LOOP)
                                                        || match(ACCEPT) )
                        {
                            /* It's a literal open bracket */
                            rval = checkcol( *yytext );
                            advance();
                            break;
                        }
                        else
                        {
                            rval = CCL;
                            if( (ccl(cclbuf)) == (char *)(long)FAILURE )
                                rval = FAILURE;
                            else if( !match( CLOSE_BRACKET ) )
                                rval = expected( CLOSE_BRACKET, 0 );
                            else
                            {
                                last_ccl = cclbuf;
                                advance();
                            }
                        }
                        break;

    default:            rval = expected( CLOSE_BRACKET, CLOSE_PAREN, COLON,
                                         OPEN_BRACKET, OPEN_PAREN, CHARACTER,
                                         NUMBER, 0 );
                        break;
    }

    if( rval != FAILURE && negative )
    {
        if( rval == CCL )                       /* reverse sense of bits */
        {
            for( p = cclbuf, i = MAXCOLS; --i >= 0 ; ++p )
                *p = !*p ;
        }
        else
        {
            memset( cclbuf, 1, MAXCOLS );       /* Create a character class */
            cclbuf[ rval ] = 0;                 /* in which all bits except */
            rval = CCL;                         /* the one corresponding to */
            last_ccl = cclbuf;                  /* rval are set.            */
        }
    }

    return rval;
}
/*----------------------------------------------------------------------*/
int     literal()
{
    int i = FAILURE;

    if( !match(OPEN_PAREN) ) {  return FAILURE;                            }
    else                     {  advance();                                 }

    if( !match(NUMBER) )     {  return expected( NUMBER, 0 );              }
    else                     {  i = checkcol( atoi( yytext ) ); advance(); }

    if( !match(CLOSE_PAREN)) {  return expected( CLOSE_PAREN, 0 );         }
    else                     {  advance();                                 }

    return i;
}
/*----------------------------------------------------------------------*/
char    *ccl( char *class )
{
    /* returns (void *)FAILURE on a syntax error, or returns a pointer to a
     * boolean array of char on success. Each true element of the array
     * corresponds to a member of the character class. The array always has
     * MAXCOLS characters in it.
     */

    int start, end, tmp;

    if( !(match(CHARACTER) || match(NUMBER) || match(OPEN_PAREN)) )
        return (char *)(long) expected( CHARACTER, NUMBER, OPEN_PAREN, 0);

    memset( class, 0, MAXCOLS );
    while( match(OPEN_PAREN) || match(NUMBER) || match(CHARACTER) )
    {
        if( (start = literal()) == FAILURE )
            if( (start = char_or_num()) == FAILURE )
            {
                expected( CHARACTER, NUMBER, OPEN_PAREN, 0 );
                return (char *)FAILURE;
            }

        if( !match(DASH) )
            end = start;
        else
        {
            advance();
            if( (end = literal()) == FAILURE )
                if( (end = char_or_num()) == FAILURE )
                {
                    expected( CHARACTER, NUMBER, OPEN_PAREN, 0 );
                    return (char *)(long)FAILURE;
                }
        }

        if( start > end )
        {
            tmp   = start;
            start = end;
            end   = tmp;
        }

        for(; start <= end; class[start++] = 1 )
            ;
    }
    return class;
}
/*----------------------------------------------------------------------*/
int     char_or_num()
{
    int rval = FAILURE;
    switch( Lookahead )
    {
    case NUMBER:        if( yytext[1] )         /* single digit */
                            error( "Must be single character, use (<num>)"
                                   "to enter binary value\n" );
                        /* fall through */

    case CLOSE_BRACKET:         /* Treat most tokens as characters in this */
    case CLOSE_PAREN:           /* context.                                */
    case COLON:
    case NOT:
    case OPEN_PAREN:
    case CHARACTER:     rval = checkcol( *yytext );
                        advance();
                        break;
    }
    return rval;
}
/*----------------------------------------------------------------------*/
void print_tables()
{
    int  i;
    char **v;

    printf("char State_table[%d][%d] =\n", Maxstate + 1, MAXCOLS );
    print_array( stdout, (int *) Statetab, Maxstate + 1, MAXCOLS );

    printf("\nchar Accepting[%d] =\n{\n", Maxstate + 1 );
    for( i = Maxstate + 1, v = Acceptab; --i >= 0; ++v )
        printf("\t%c%c\t/* %3d */\n",
                        *v                      ? '1' : '0' ,
                        v < &Acceptab[Maxstate] ? ',' : ' ' ,
                        v - Acceptab );
    printf("};\n\n");

    printf(        "action( int state_num )\n"
                   "{\n"
                   "    switch( state_num )\n"
                   "    {\n"                           );

    for( i = Maxstate + 1, v = Acceptab; --i >= 0; ++v )
        if( *v )
            printf("    case %d:\n"
                   "            %s\n"
                   "            break;\n", v - Acceptab, *v );

    printf(        "    default: return -1;\n"
                   "    }\n"
                   "}\n"                       );
}

/*======================================================================*/

main()
{
    /* Initialize state table to -1's, fill it, and print it. The table is
     * initialized one row at a time because sizeof(Statetab) might not
     * fit in an int.
     */

    int i;
    int (*p)[MAXCOLS];        /* Pointer to a MAXCOLS-element array of ints.*/
                              /* Incrementing p skips past an entire array. */

    for( i = MAXSTATES, p = Statetab; --i >= 0; )
        memset( p++, -1, MAXCOLS * sizeof(int) );

    description();

    if( !Error_count )
        print_tables();

    exit( Error_count );
}